#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h> 

#include "pico/stdlib.h"
#include "hardware/pwm.h"
#include "hardware/adc.h"

#include "EA_TFT.h"
#include "EA_TOUCH.h"
#include "Screen.h"

#define FWVER          110   // = V1.1

//-------------------------------------------------------------------------------------------------

#define TFT_PWM   9  // Pin 12

typedef struct
{ int16_t zs;
  int16_t sec;
  int16_t screen;
  int16_t screenold;
  int16_t send;
  int16_t tftmodul;
  int16_t tftangle;
  int16_t tfttouch;
  int16_t dip;
  int16_t dipnew;
} EVENT;

//-------------------------------------------------------------------------------------------------

#define BUFSIZE (225*1024+128)

//uint8_t DIPx[8] = {13,12,11,10, 18,19,20,21};   // Prototyp: GPIO DIP1.8 
uint8_t  DIPx[8] = {10,11,12,13, 18,19,20,21}; // GPIO DIP1.8 
uint     gSlicenum, gChannel, gTFTmess;
uint16_t gPWMcountmax, gLED;
float    gVled;

uint8_t BMPbuffer[BUFSIZE];

//-------------------------------------------------------------------------------------------------

#define TFTMODUL       (event.tfttouch<<7 | gTFTAngle/90<<5 | gTFTmodul)
#define USBscreen(nr)   if(event.send) USBprintf("\x1BS%c%c%c", FWVER, TFTMODUL, nr)

void USBprintf(char *format, ...)
{
  va_list parg;
  char   *cp, str[128];
  int     len;

  va_start(parg, format);
  len = vsprintf(str, format, parg);
  va_end(parg);

  if(stdio_usb_connected()  &&  len > 0)
   { cp = str;
     while(len--)
        putchar(*cp++);
   }
}

void setLED(uint16_t percent, int16_t TFTmodul)
{
  uint LEDmax = 0;

  gLED = percent;

  switch(TFTmodul)
   { case EA_TFT009:  LEDmax = 20;  break;
     case EA_TFT015:  LEDmax = 80;  break;
     case EA_TFT020:  LEDmax = 20;  break;
     case EA_TFT028:  LEDmax = 20;  break;
     case EA_TFT035:  LEDmax = 40;  break;
     case EA_TFT024:  LEDmax = 60;  break;
   }

  pwm_set_chan_level(gSlicenum, gChannel,  (uint)gPWMcountmax * LEDmax/80 * percent/100);
}

void getLEDmax(void)
{
  gTFTmess = 0;
  setLED(100, EA_TFT015); // for measure set to max. LED current (80mA)

 // LED PWM 100% = 80mA 
 // offen  33.0V (alte Bestückung)
 // TFT009  9.7V (9.7V)    // max. 20mA;  1 path with 1 LED and 75 Ohm in series
 // TFT015  3.8V (3.3V)    // max. 80mA;  4 path with 1 LED 
 // TFT035 20.3V (19.1V)   // max. 40mA;  2 path with 6 LED
 // TFT020 25.5V (23.8V)   // max. 20mA;  1 path with 8 LED
 // TFT028 27.1V (25.0V)   // max. 20mA;  1 path with 9 LED

  sleep_ms(30);
  gVled = 33.0 * (float)adc_read() / 4096.0;
  if(gVled > 2.5  &&  gVled < 30.0)
   { if(gVled < 5.0)        gTFTmess = EA_TFT015;
     else if(gVled < 12.0)  gTFTmess = EA_TFT009;
     else if(gVled < 22.0)  gTFTmess = EA_TFT035;
     else if(gVled < 26.0)  gTFTmess = EA_TFT020;
     else                   gTFTmess = EA_TFT028;
   }

  setLED(100, gTFTmess);
}

uint8_t getDIP(void)
{
  uint8_t dip = 0;

  if(gpio_get(DIPx[0]))  dip |= 0x01;
  if(gpio_get(DIPx[1]))  dip |= 0x02;
  if(gpio_get(DIPx[2]))  dip |= 0x04;
  if(gpio_get(DIPx[3]))  dip |= 0x08;
  if(gpio_get(DIPx[4]))  dip |= 0x10;
  if(gpio_get(DIPx[5]))  dip |= 0x20;
  if(gpio_get(DIPx[6]))  dip |= 0x40;
  if(gpio_get(DIPx[7]))  dip |= 0x80;

  return(dip ^ 0xFF);
}

void getDIPevent(EVENT *e)
{
  uint8_t dip = getDIP();

  switch(dip & 0x3F)
   { case 0x01:  e->tftmodul = EA_TFT009;  break;
     case 0x02:  e->tftmodul = EA_TFT015;  break;
     case 0x04:  e->tftmodul = EA_TFT020;  break;
     case 0x08:  e->tftmodul = EA_TFT028;  break;
     case 0x10:  e->tftmodul = EA_TFT035;  break;
     case 0x20:  e->tftmodul = EA_TFT024;  break;
     default:    e->tftmodul = gTFTmess;
   }
 
  switch(dip >> 6)
   { case 0:  e->tftangle =  90;  break;
     case 1:  e->tftangle = 270;  break;
     case 2:  e->tftangle = 180;  break;
     case 3:  e->tftangle =   0;  break;
   }

  e->dipnew = (dip != e->dip);
  e->dip    = dip;
}

bool repeating_timer_callback(struct repeating_timer *t)
{
  EVENT *e = (EVENT *)t->user_data;

  getDIPevent(e);

  if(++e->zs == 10)
   { e->zs = 0;
     e->sec++;
     if(e->screen >= 1  &&  e->screen < 4  &&  e->sec % 3 == 0)         // alle 3 Sekunden einen Screenwechsel
      { if(++e->screen == 4)
           e->screen = 1;
      }
   }

  return true;
}

void nextchar(uint8_t **ppdat)
{
  uint8_t *cp = *ppdat;

  while(*cp != 0)
   { if(*cp == ';'  ||  *cp == '$'  ||  *cp == '%'  ||  isdigit(*cp)  ||  *cp == '\"'  ||  *cp == '\'')
        break;
     cp++;
   }

  *ppdat = cp;
}

uint16_t getvalue(uint8_t **ppdat)
{
  uint8_t *cp = *ppdat;
  int     wert;

  if(*cp == '$')
   { sscanf(++cp, "%x", &wert);
     while(isxdigit(*cp))
        cp++;
   }
  else if(*cp == '%')
   { cp++;
     wert = 0;
     while(*cp == '0'  ||  *cp == '1')
        wert = (wert<<1) + (*cp++ - '0');
   }
  else if(isdigit(*cp))
   { sscanf(cp, "%d", &wert);
     while(isdigit(*cp))
        cp++;
   }

  *ppdat = cp;

 return wert;
}

int16_t getnextvalue(uint8_t **ppdat)
{
  int16_t wert;
   
  nextchar(ppdat);
  wert = getvalue(ppdat);

  return wert;
}

int getUntilCharOrLF(uint8_t *pdat, uint8_t end)
{
  int c;

  while(1)
   { c = getchar();
     if(c == end  ||  c == 0x0A)
        break;
    *pdat++ = c;
   }

  *pdat = 0;

  return c;
}

int16_t getString(uint8_t *cp, uint8_t *str)
{
  int16_t len = 0;
  uint8_t end;

  while(*cp != 0)
   { nextchar(&cp);
     if(*cp == ';')
        break;
     if(isdigit(*cp)  ||  *cp == '$'  ||  *cp == '%')
        str[len++] = getnextvalue(&cp);
     else if(*cp == '\"'  ||  *cp == '\'')
      { end = *cp++;
        while(*cp >= 32  &&  *cp != end)
           str[len++] = *cp++;
        if(*cp == end)
           cp++;
      }
   }
   str[len] = 0;

   return len;
}

int getBMPdata(uint8_t *pdat, int32_t maxsize)
{
  int32_t rc, size;
  
 *pdat++ = 'B';
 *pdat++ = 'M';

  rc     = getchar();
 *pdat++ = rc;
  size   = rc;
  rc     = getchar();
 *pdat++ = rc;
  size  |= rc << 8;
  rc     = getchar();
 *pdat++ = rc;
  size  |= rc << 16;
  rc     = getchar();
 *pdat++ = rc;
  size |= rc << 24;

  if(size <= maxsize)
   { size -= 6;
     while(size > 0)
      { if((rc=getblock_timeout_us(pdat, 512, 100000)) == PICO_ERROR_TIMEOUT)
           break;
        pdat += rc;
        size -= rc;
      }
   }

  return (size == 0);    
}

int main()
{
   struct repeating_timer timer;
   EVENT    event;
   int32_t  rc, size;
   uint8_t *pdat, befchar1, befchar2, endchar;
   uint16_t i, x,y, x2,y2, r,g,b, r2,g2,b2;

   stdio_init_all();
 
   for(i=0; i < 8; i++)
    { gpio_init(DIPx[i]);
      gpio_set_dir(DIPx[i], GPIO_IN);
      gpio_pull_up(DIPx[i]);
    }

   adc_init();
   adc_gpio_init(26);
   adc_select_input(0);

   gpio_set_function(TFT_PWM, GPIO_FUNC_PWM);
   gSlicenum = pwm_gpio_to_slice_num(TFT_PWM);
   gChannel  = pwm_gpio_to_channel(TFT_PWM);
   pwm_set_clkdiv(gSlicenum, 125);       // 125 MHz / 125 = 1MHz clk for PWM-Timer
   gPWMcountmax = 1000000/95;            // 95Hz
   pwm_set_wrap(gSlicenum, gPWMcountmax);
   pwm_set_chan_level(gSlicenum, gChannel, gPWMcountmax/4);   // Iled = max.LED 80mA/4 = 20mA 
   pwm_set_enabled(gSlicenum, true);

   getLEDmax();

   memset(&event, 0, sizeof(event));
   getDIPevent(&event);
   event.dipnew = 0;
   event.screen = (event.tftmodul != 0);

   TFTspi_init(event.tftmodul, 16000000);
   setTFTangle(event.tftangle);
   TOUCHi2c_init(400000);
   event.tfttouch = check_Touch();

   add_repeating_timer_ms(100, repeating_timer_callback, &event, &timer);
    
   while(1)
    { 
      if(event.dipnew)
       { event.dipnew = 0;
         if(event.tftmodul != gTFTmodul)
          { if(event.tftmodul == EA_TFT035  ||  gTFTmodul == EA_TFT035) 
               TFTspi_init(event.tftmodul, 16000000);
            else
               TFTinit(event.tftmodul);
            setLED(gLED, gTFTmodul);
            TOUCHi2c_init(400000);
            event.tfttouch = check_Touch();
          }
         setTFTangle(event.tftangle);
         event.screen = 1;  event.screenold = 0;  
         event.sec = 0;
       }

      if(event.screen  &&  event.screen != event.screenold)
       { USBscreen(event.screen);
         Screen(event.screen);
         event.screenold = event.screen;
       }

      if(gTouchOK)
       { //if(event.screen > 0  &&  event.screen < 255)
         if(event.screen < 255)
          { event.screen = 255;  // 255 = ClrTouchScreenDraw
            USBscreen(event.screen);
            Screen(event.screen);
          }
         else if(gX < 40  &&  gY < 16)
          { if(gTouchEvent == TOUCH_UP)
             { Screen(255);  // 255 = ClrTouchScreenDraw
               USBscreen(255);
             }
          }
         else if(gX > gTFTwidth-32  &&  gY < 16)
          { if(gTouchEvent == TOUCH_UP)
             { event.screen = 1;  event.screenold = 0;  
               event.sec = 0;
             }
          }
         else
          { switch(gTouchEvent)
             { case TOUCH_DOWN: TFTfill(GREEN, gX-1,gY-1, gX+1,gY+1);  break;
               case TOUCH_DRAG: TFTfill(WHITE, gX,gY, gX,gY);          break;
               case TOUCH_UP:   TFTfill(RED,   gX-1,gY-1, gX+1,gY+1);  break;
             }
            USBprintf("\x1BT%c%c%c%c%c%c%c", FWVER, TFTMODUL, gTouchEvent, LO(gX),HI(gX), LO(gY),HI(gY));
          }
         gTouchOK = 0;
       }

      if(stdio_usb_connected())
       { if((rc=getchar_timeout_us(0)) != PICO_ERROR_TIMEOUT)
          { if(rc == 0x1B)
             { befchar1 = getchar();
               if(befchar1 >= 'a'  &&  befchar1 <= 'z')  befchar1 &= 0xDF;
               pdat = BMPbuffer;
               getUntilCharOrLF(pdat, 0x0A);
               switch(befchar1)
                { case '?':  event.send = 1;
                             USBscreen(event.screen);
                          break;
                  case 'D':  i = getnextvalue(&pdat);
                             if(i == 0)
                                event.screen = 0;
                             else if(i >= 1  &&  i <= 3)
                              { event.screen = i;  event.screenold = i; 
                                event.sec = 0;
                                USBscreen(event.screen);
                                Screen(event.screen);
                              }
                          break;
                  case 'S':  i = getnextvalue(&pdat);
                             if(i >= 1  &&  i <= 3)
                              { event.screen = 0;
                                Screen(i);
                              }
                          break;
                  case 'P':  i = getnextvalue(&pdat);
                             if(i >= 50  &&  i <= 5000)
                              { gPWMcountmax =  (int)1000000/i;
                                pwm_set_wrap(gSlicenum, gPWMcountmax);
                                setLED(gLED, gTFTmodul);
                              }
                          break;
                  case 'V':  USBprintf(" %4.1fV", gVled);
                          break;
                }
             }
            else if(rc == '#')
             { befchar1 = getchar() & 0xDF;
               befchar2 = getchar() & 0xDF;
               pdat     = BMPbuffer;
               endchar  = (befchar1=='B' && befchar2=='M') ? 'M' : 0x0A;
               endchar  = getUntilCharOrLF(pdat, endchar);
               switch(befchar1)
                { case 'L':  switch(befchar2)
                              { case 'B': i = getnextvalue(&pdat);
                                          setLED(i, gTFTmodul);
                                        break;
                              }
                          break;
                  case 'D':  switch(befchar2)
                              { case 'O': i = getnextvalue(&pdat);
                                          setTFTangle(i);
                                        break;
                                case 'C': TFTclear(BLACK);
                                          event.screen = 0;
                                        break;
                                case 'F': r = getnextvalue(&pdat);
                                          g = getnextvalue(&pdat);
                                          b = getnextvalue(&pdat);
                                          TFTclear(RGB565(r,g,b));
                                          event.screen = 0;
                                        break;
                              }
                          break;
                  case 'F':  switch(befchar2)
                              { case 'A': r  = getnextvalue(&pdat);
                                          g  = getnextvalue(&pdat);
                                          b  = getnextvalue(&pdat);
                                          x  = getnextvalue(&pdat);
                                          y  = getnextvalue(&pdat);
                                          x2 = getnextvalue(&pdat);
                                          y2 = getnextvalue(&pdat);
                                          TFTfill(RGB565(r,g,b), x,y, x2,y2);
                                          event.screen = 0;
                                        break;
                              }
                          break;
                  case 'T':  switch(befchar2)
                              { case 'C': r  = getnextvalue(&pdat);
                                          g  = getnextvalue(&pdat);
                                          b  = getnextvalue(&pdat);
                                          r2 = getnextvalue(&pdat);
                                          g2 = getnextvalue(&pdat);
                                          b2 = getnextvalue(&pdat);
                                          setTFTcolor(RGB565(r,g,b), RGB565(r2,g2,b2));
                                        break;
                                case 'F': i = getnextvalue(&pdat);
                                          x = getnextvalue(&pdat);
                                          y = getnextvalue(&pdat);
                                          setTFTfont(i, x,y);
                                        break;
                                case 'P': x = getnextvalue(&pdat);
                                          y = getnextvalue(&pdat);
                                          if(getString(pdat, BMPbuffer+1000) > 0)
                                           { TFTtext(x,y, BMPbuffer+1000);
                                             event.screen = 0;
                                           }
                                        break;
                              }
                          break;
                  case 'B':  switch(befchar2)
                              { case 'M': x = getnextvalue(&pdat);
                                          y = getnextvalue(&pdat);
                                          while(*pdat != 0  &&  *pdat != 'B'  &&  *pdat != '\"'  &&  *pdat != '\''  &&  *pdat != ';')
                                             pdat++;
                                          if(*pdat == 'B'  &&  endchar == 'M')
                                           { if(getBMPdata(BMPbuffer, BUFSIZE))
                                              { TFTbitmap(x,y, BMPbuffer, 0,0);
                                                event.screen = 0;
                                              }
                                           }
                                        break;
                              }
                          break;
                }
             }
            else if(rc == 'B')
             { if(getchar_timeout_us(100000) == 'M')
                { if(getBMPdata(BMPbuffer, BUFSIZE))
                   { TFTbitmap(0,0, BMPbuffer, 0,0);
                     event.screen = 0;
                   }
                }
             }
            else if(rc == ';')
               getUntilCharOrLF(pdat, 0x0A);
          }
       }
   }

  return 0;
}
